#lang slideshow
;; This Racket program creates a slideshow of randomly created
;; "mathematical images", images created with a nested math
;; expression according to a grammar like this one:
;;
;;   EXPR = x |
;;          y |
;;          (sinpi EXPR) |
;;          (cospi EXPR) |
;;          (* EXPR EXPR) |
;;          (avg EXPR EXPR)
;;
;; Inspired by this thread:
;; http://www.mail-archive.com/users@racket-lang.org/msg06079.html

;; TODO:
;;   * Make it possible to inspect which expression
;;     generated a given image.
;;   * Save high-resolution images (for wallpapers)

; You may want to change the width and
; height of the generated images.
; Note that images are scaled (up or down)
; to fit in the screen.
(define width 160)
(define height 120)

;; -----------------------------------------------------------------------------
(require (only-in picturing-programs
                  build3-image
                  real->int)
         racket/generator)

;; -----------------------------------------------------------------------------
;; Slideshow config

;(set-page-numbers-visible! #f)
(set-margin! 0)

(define scale/x (/ client-w width))
(define scale/y (/ client-h height))

;; -----------------------------------------------------------------------------
;; Image generation code

;; ???
(define (color-from-expression expr)
  (let ([x-interpolator (range-interpolator 0 (sub1 width) -1 1)]
        [y-interpolator (range-interpolator 0 (sub1 height) -1 1)]
        [output-interpolator (range-interpolator -1 1 0 255)])
    (λ (x y)
      (real->int
       (output-interpolator
        (expr (x-interpolator x)
              (y-interpolator y)))))))

;; N -> EXPR
;; Generate a mathematical expression of depth i
(define (generate-expression i)
  (if (> i 1)
      (let ([sub-expr (generate-expression (sub1 i))])
        (case (random 2)
          [(0) (case (random 2)
                 [(0) (lambda (x y) (sinpi (sub-expr x y)))]
                 [(1) (lambda (x y) (cospi (sub-expr x y)))])]
          [(1) (let([sub-expr2 (generate-expression (sub1 i))])
                 (case (random 2)
                   [(0) (lambda (x y)
                          (* (sub-expr x y)
                             (sub-expr2 x y)))]
                   [(1) (lambda (x y)
                          (avg (sub-expr x y)
                               (sub-expr2 x y)))]))]))
      (case (random 2)
        [(0) (lambda (x y) x)]
        [(1) (lambda (x y) y)])))

;; ???
(define (range-interpolator from-min from-max to-min to-max)
  (λ (x)
    (+ to-min
       (* (/ (- to-max to-min)
             (- from-max from-min))
          (- x
             from-min)))))

;; Helper functions to build mathematical expressions

(define (avg x y)
  (/ (+ x y)
     2))

(define (sinpi x)
  (sin (* pi x)))

(define (cospi x)
  (cos (* pi x)))

;; -----------------------------------------------------------------------------
;; Slideshow generation

(define (generate-slide n image)
  (slide
   ;(cc-superimpose
   (scale
    (bitmap
     (send
      image
      get-bitmap))
    scale/x scale/y)
   ;(caption n))
   ))

(define (caption n)
  (let ([text (bt (format "~a" n))])
    (colorize (scale/improve-new-text text (/ (* 0.8 client-h)
                                              (pict-height text)))
              "white")))

;; -----------------------------------------------------------------------------
;; Create a list of images and put them on a slideshow

(define image-generator
  (infinite-generator
   (yield
    (build3-image width height
                  (color-from-expression (generate-expression 5))
                  (color-from-expression (generate-expression 6))
                  (color-from-expression (generate-expression 7))))))

(for ([i (in-range 1 11)]
      [img (in-producer image-generator #f)])
  (printf "Creating slide #~a~n" i)
  (generate-slide i img))

; -------------- "interesting image" ------------------

(define expr
  (λ (x y)
    (cospi (* (avg (sinpi y) x) (cospi (* x y))))))

(let ([cfe (color-from-expression expr)])
  (build3-image
   width height
   cfe cfe cfe))